"""Pylint integration to Pylama."""
import logging
from os import path as op, environ

from astroid import MANAGER
from pylama.lint import Linter as BaseLinter
from pylint.__pkginfo__ import numversion
from pylint.lint import Run
from pylint.reporters import BaseReporter


CURDIR = op.abspath(op.dirname(__file__))
HOME_RCFILE = op.abspath(op.join(environ.get('HOME', ''), '.pylintrc'))
LAMA_RCFILE = op.abspath(op.join(CURDIR, 'pylint.rc'))


logger = logging.getLogger('pylama')


class Linter(BaseLinter):
    """Check code with Pylint."""

    @staticmethod
    def run(path, code, params=None, ignore=None, select=None, **meta):
        """Pylint code checking.

        :return list: List of errors.
        """
        logger.debug('Start pylint')

        clear_cache = params.pop('clear_cache', False)
        if clear_cache:
            MANAGER.astroid_cache.clear()

        class Reporter(BaseReporter):

            def __init__(self):
                self.errors = []
                super(Reporter, self).__init__()

            def _display(self, layout):
                pass

            def handle_message(self, msg):
                self.errors.append(dict(
                    lnum=msg.line,
                    col=msg.column,
                    text="%s %s" % (msg.msg_id, msg.msg),
                    type=msg.msg_id[0]
                ))

        params = _Params(ignore=ignore, select=select, params=params)
        logger.debug(params)

        reporter = Reporter()

        kwargs = {
            (numversion[0] == 1 and 'exit' or 'do_exit'): False
        }

        Run([path] + params.to_attrs(), reporter=reporter, **kwargs)

        return reporter.errors


class _Params(object):
    """Store pylint params."""

    def __init__(self, select=None, ignore=None, params=None):

        params = dict(params)

        if op.exists(HOME_RCFILE):
            params['rcfile'] = HOME_RCFILE

        if select:
            enable = params.get('enable', None)
            params['enable'] = select | set(enable.split(",") if enable else [])

        if ignore:
            disable = params.get('disable', None)
            params['disable'] = ignore | set(disable.split(",") if disable else [])

        self.params = dict(
            (name.replace('_', '-'), self.prepare_value(value))
            for name, value in params.items() if value is not None)

    @staticmethod
    def prepare_value(value):
        """Prepare value to pylint."""
        if isinstance(value, (list, tuple, set)):
            return ",".join(value)

        if isinstance(value, bool):
            return "y" if value else "n"

        return str(value)

    def to_attrs(self):
        """Convert to argument list."""
        return ["--%s=%s" % item for item in self.params.items()]

    def __str__(self):
        return " ".join(self.to_attrs())

    def __repr__(self):
        return "<Pylint %s>" % self

# pylama:ignore=W0403



"""MyPy support."""

from mypy import api

from pylama.lint import Linter as Abstract


class _MyPyMessage(object):
    """Parser for a single MyPy output line."""
    types = {
        'error': 'E',
        'warning': 'W',
        'note': 'N'
    }

    def __init__(self, line):
        self.filename = None
        self.line_num = None
        self.column = None
        self.text = None
        self.note = None
        self.message_type = None
        self.valid = False

        self._parse(line)

    def _parse(self, line):
        """Parse the output line"""
        try:
            result = line.split(':', maxsplit=4)
            filename, line_num_txt, column_txt, message_type, text = result
        except ValueError:
            return

        try:
            self.line_num = int(line_num_txt.strip())
            self.column = int(column_txt.strip())
        except ValueError:
            return

        self.filename = filename
        self.message_type = message_type.strip()
        self.text = text.strip()
        self.valid = True

    def add_note(self, note):
        """Add in additional information about this message"""
        self.note = note

    def to_result(self):
        """Convert to the Linter.run return value"""
        text = [self.text]
        if self.note:
            text.append(self.note)

        return {
            'lnum': self.line_num,
            'col': self.column,
            'text': ' - '.join(text),
            'type': self.types.get(self.message_type, '')
        }


class Linter(Abstract):
    """MyPy runner."""

    @staticmethod
    def run(path, code=None, params=None, **meta):
        """Check code with mypy.

        :return list: List of errors.
        """
        args = [path, '--follow-imports=skip', '--show-column-numbers']
        stdout, stderr, status = api.run(args)
        messages = []
        for line in stdout.split('\n'):
            line.strip()
            if not line:
                continue
            message = _MyPyMessage(line)
            if message.valid:
                if message.message_type == 'note':
                    if messages[-1].line_num == message.line_num:
                        messages[-1].add_note(message.text)
                else:
                    messages.append(message)

        return [m.to_result() for m in messages]
